import bpy
from . import server as DS

class DISPLAYSERVER_PT_panel(bpy.types.Panel):
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "DISPLAYSERVER"
    bl_label = "Display Server"

    def addButton(self, row, buttonClazz: bpy.types.Operator):
        row.operator(buttonClazz.bl_idname, text=buttonClazz.bl_label)

    def draw(self, context):
        layout = self.layout

        box = layout.box()
        box.label(text="Main Controls")

        row = box.row()
        self.addButton(row, DISPLAYSERVER_OT_CurCol)
        self.addButton(row, DISPLAYSERVER_OT_Start)

        row = box.row()
        self.addButton(row, DISPLAYSERVER_OT_ReportStatus)

        #last message text field
        box.prop(context.scene, "DS_recent_message_text")
        # color checkbox
        box.prop(context.scene, "DS_is_using_one_color")
        # uncovering checkbox
        box.prop(context.scene, "DS_is_creating_uncovering_tree")
        # flexi checkbox
        box.prop(context.scene, "DS_is_doing_flexi_lines")

        row = box.row()
        self.addButton(row, DISPLAYSERVER_OT_HideOne)
        self.addButton(row, DISPLAYSERVER_OT_ClearOne)

        row = box.row()
        self.addButton(row, DISPLAYSERVER_OT_Start)
        self.addButton(row, DISPLAYSERVER_OT_Stop)


def updateOneColorCBox(self, context):
    DS = getSERV()
    if DS is not None:
        DS.is_using_one_color = not self.DS_is_using_one_color

def updateUncoveringCBox(self, context):
    DS = getSERV()
    if DS is not None:
        DS.is_creating_uncovering_tree = self.DS_is_creating_uncovering_tree

def updateFlexiCBox(self, context):
    DS = getSERV()
    if DS is not None:
        DS.is_doing_flexi_lines = self.DS_is_doing_flexi_lines

def updateRecentMsgField(message: str):
    bpy.context.scene.DS_recent_message_text = message


def getSERV():
    # create the server on the first access
    # NB: Blender cannot start this server while register() below
    # NB: so starting is done here, when already in fully-started context
    if bpy.types.Scene.DISPLAYSERVER is None:
        bpy.types.Scene.DISPLAYSERVER = DS.DisplayGraphicsServer("Blender",9081)

    return bpy.types.Scene.DISPLAYSERVER

# -------------------------------------------------

class DISPLAYSERVER_OT_ReportStatus(bpy.types.Operator):
    """Reports the current status and settings of the DisplayServer."""
    bl_idname = "ds.report_status"
    bl_label = "Report Current Status"

    def execute(self, context):
        DISPLAYSERVER = getSERV()
        if DISPLAYSERVER is None:
            self.report({'ERROR'},"Couldn't find any DisplayServer.")
            return {'FINISHED'}

        print(f"Server at {DISPLAYSERVER.serverName}:{DISPLAYSERVER.serverPort} is {DISPLAYSERVER.serv_status}.")
        print(f"When fully started, it processes data every {DISPLAYSERVER.blender_auto_updates_per_second} seconds.")
        colName = DISPLAYSERVER.refCollection.name if DISPLAYSERVER.refCollection is not None else "_not_set_"
        print(f"It is currently hooked to {DISPLAYSERVER.refName} in {colName} master collection.")
        print("Next time lineage comes, it will "
            +("not " if DISPLAYSERVER.is_using_one_color else "")
            +"use embryo colors, and will show "
            +("unfolding" if DISPLAYSERVER.is_creating_uncovering_tree else "complete")
            +" lineage tree with "
            +("flexible" if DISPLAYSERVER.is_doing_flexi_lines else "fixed")
            +" edges.")
        return {'FINISHED'}

class DISPLAYSERVER_OT_CurCol(bpy.types.Operator):
    """Sets the DisplayServer to use the currently active Blender collection."""
    bl_idname = "ds.current_collection"
    bl_label = "Set Collection"

    def execute(self, context):
        DISPLAYSERVER = getSERV()
        if DISPLAYSERVER is None:
            self.report({'ERROR'},"Couldn't find any DisplayServer.")
            return {'FINISHED'}

        DISPLAYSERVER.switch_to_master_collection( bpy.context.collection )

        print(f"Collection set to {DISPLAYSERVER.refCollection.name}")
        return {'FINISHED'}

class DISPLAYSERVER_OT_Start(bpy.types.Operator):
    """(Re)Starts the DisplayServer using the current settings."""
    bl_idname = "ds.start"
    bl_label = "Start Server"

    def execute(self, context):
        DISPLAYSERVER = getSERV()
        if DISPLAYSERVER is None:
            self.report({'ERROR'},"Couldn't find any DisplayServer.")
            return {'FINISHED'}
        if DISPLAYSERVER.refCollection is None:
            self.report({'ERROR'},"Please \"Set Collection\" first.")
            return {'FINISHED'}

        DISPLAYSERVER.last_message_callback = updateRecentMsgField
        DISPLAYSERVER.start()

        # side job: enlarge Blender's view depth
        for a in bpy.context.screen.areas:
            if a.type == 'VIEW_3D':
                a.spaces.active.clip_end = 100000
                break
        return {'FINISHED'}

class DISPLAYSERVER_OT_Stop(bpy.types.Operator):
    """Stops the DisplayServer: will continue to exist, only will not be listening."""
    bl_idname = "ds.stop"
    bl_label = "Stop Server"

    def execute(self, context):
        DISPLAYSERVER = getSERV()
        if DISPLAYSERVER is None:
            self.report({'ERROR'},"Couldn't find any DisplayServer.")
            return {'FINISHED'}

        DISPLAYSERVER.stop()
        return {'FINISHED'}

class DISPLAYSERVER_OT_HideOne(bpy.types.Operator):
    """Hides (not deletes) recently displayed objects (hiding is performance friendly)."""
    bl_idname = "ds.hide_recent"
    bl_label = "Hide Recent"

    def execute(self, context):
        DISPLAYSERVER = getSERV()
        if DISPLAYSERVER is None:
            self.report({'ERROR'},"Couldn't find any DisplayServer.")
            return {'FINISHED'}

        print(f"Hiding in {DISPLAYSERVER.refName} in {DISPLAYSERVER.refCollection.name} collection")
        DISPLAYSERVER.hide_current_content_all()
        return {'FINISHED'}

class DISPLAYSERVER_OT_ClearOne(bpy.types.Operator):
    """Deletes (not hides) recently displayed objects."""
    bl_idname = "ds.clear_recent"
    bl_label = "Clear Recent"

    def execute(self, context):
        DISPLAYSERVER = getSERV()
        if DISPLAYSERVER is None:
            self.report({'ERROR'},"Couldn't find any DisplayServer.")
            return {'FINISHED'}

        print(f"Clearing in {DISPLAYSERVER.refName} in {DISPLAYSERVER.refCollection.name} collection")
        DISPLAYSERVER.clear_current_content_all()
        return {'FINISHED'}

# -------------------------------------------------

classes = (
    DISPLAYSERVER_PT_panel,
    DISPLAYSERVER_OT_CurCol,
    DISPLAYSERVER_OT_ReportStatus,
    DISPLAYSERVER_OT_HideOne,
    DISPLAYSERVER_OT_ClearOne,
    DISPLAYSERVER_OT_Start,
    DISPLAYSERVER_OT_Stop,
)

def register():
    scene = bpy.types.Scene

    # properties
    scene.DS_recent_message_text = bpy.props.StringProperty(name="Recent lineage:")
    scene.DS_is_using_one_color = bpy.props.BoolProperty(name="Next time use embryo colors", default=True, update=updateOneColorCBox)
    scene.DS_is_creating_uncovering_tree = bpy.props.BoolProperty(name="Next time uncovering lineage", default=False, update=updateUncoveringCBox)
    scene.DS_is_doing_flexi_lines = bpy.props.BoolProperty(name="Next time do flexi edges", default=False, update=updateFlexiCBox)
    scene.DISPLAYSERVER = None
    # NB: DISPLAYSERVER is created later! in getSERV()

    # classes
    for c in classes:
        bpy.utils.register_class(c)


def unregister():
    scene = bpy.types.Scene

    # properties
    del scene.DS_recent_message_text
    del scene.DS_is_using_one_color
    del scene.DS_is_creating_uncovering_tree
    del scene.DS_is_doing_flexi_lines
    del scene.DISPLAYSERVER

    # classes
    for c in reversed(classes):
        bpy.utils.unregister_class(c)
