from grpc import server, RpcError
from concurrent import futures
from time import sleep
import DisplayServer.points_and_lines_pb2 as points_and_lines_pb2
import DisplayServer.points_and_lines_pb2_grpc as points_and_lines_pb2_grpc
import bpy
import mathutils
from . import lines_straight


class DisplayGraphicsServer(points_and_lines_pb2_grpc.PointsAndLinesServicer):
    def __init__(self, serverName, serverPort):
        self.serverName = serverName
        self.serverPort = serverPort

        self.refMat = bpy.data.materials.new(self.serverName+" default color")
        self.refMat.diffuse_color = [0,1,0,1]
        self.refMat.roughness = 10

        self.init_serv()
        print("Server "+self.serverName+" initiated")

    # starts up networking
    def init_serv(self):
        self.serv = server( futures.ThreadPoolExecutor(4,self.serverName) )
        points_and_lines_pb2_grpc.add_PointsAndLinesServicer_to_server(self,self.serv)
        self.serv.add_insecure_port('[::]:%d'%self.serverPort)
        self.serv_status = "initiated"

    # create or re-use (canonical) reference objects
    def init_refObjs(self):
        name = self.serverName+" ref sphere"
        # try to reuse the reference objects if they do exist already
        self.refSphere = self.refCollection.objects.get(name)
        if self.refSphere is None:
            bpy.ops.mesh.primitive_uv_sphere_add(radius=1)
            self.refSphere = bpy.context.active_object
            self.refSphere.name = name
            self.refSphere.hide_viewport = True
        self.refSphere["entity_type"]="lineage"
        self.refSphere["shape"]='B'

        name = self.serverName+" ref ellipsoid"
        self.refEllipsoid = self.refCollection.objects.get(name)
        if self.refEllipsoid is None:
            bpy.ops.mesh.primitive_uv_sphere_add(radius=1)
            self.refEllipsoid = bpy.context.active_object
            self.refEllipsoid.name = name
            self.refEllipsoid.hide_viewport = True
        self.refEllipsoid["entity_type"]="lineage"
        self.refEllipsoid["shape"]='E'

        name = self.serverName+" ref cylinder"
        self.refCylinder = self.refCollection.objects.get(name)
        if self.refCylinder is None:
            bpy.ops.mesh.primitive_cylinder_add(radius=1, depth=2)
            self.refCylinder = bpy.context.active_object
            self.refCylinder.name = name
            self.refCylinder.hide_viewport = True
        self.refCylinder["entity_type"]="lineage"
        self.refCollection["shape"]='L'

        name = self.serverName+" ref empty"
        self.refEmpty = self.refCollection.objects.get(name)
        if self.refEmpty is None:
            bpy.ops.object.empty_add(type='PLAIN_AXES', align='WORLD', location=(0, 0, 0), scale=(1, 1, 1))
            self.refEmpty = bpy.context.active_object
            self.refEmpty.name = name
            self.refEmpty.hide_viewport = True

        name = self.serverName+" ref text"
        self.refText = self.refCollection.objects.get(name)
        if self.refText is None:
            bpy.ops.object.text_add(enter_editmode=False, align='WORLD', location=(0, 0, 0), scale=(1, 1, 1))
            self.refText = bpy.context.active_object
            self.refText.name = name
            self.refText.hide_viewport = True


    # (re)sets reference attributes
    def switch_to_master_collection(self, refCollection):
        self.refCollection = refCollection
        self.refCurrentCollection = None
        self.refCurrentNode = None
        self.refName = "default_parent"

    def duplicate_ref_material(self):
        self.refMat = self.refMat.copy()
        self.refMat.name = self.serverName+" default color for "+self.refName

    def set_ref_material(self):
        self.refMat = bpy.data.materials.get(self.serverName+" default color for "+self.refName)


    # returns 0 if started, or -1 if the server is not yet fully initialized
    def start(self):
        if self.refCollection is None:
            print("Server is not starting: Not all attribs are already initialized")
            return -1

        if self.refSphere is None or self.refEllipsoid is None or self.refCylinder is None \
                or self.refEmpty is None or self.refText is None:
            self.init_refObjs()

        # skip if it is (hopefully) properly running
        if self.serv_status.startswith("listening"):
            print("hmm... records say it is running happily, so I'm skipping now")
            print("[ if it is broken, rescue it with a call to init_serv() and then start() ]")
            return

        # re-create the serv instance if it once existed and was stopped
        if self.serv_status.startswith("stopped"):
            self.init_serv()

        self.serv.start()
        self.serv_status = "listening"
        print("Server "+self.serverName+" started")

        # re-reset internal data, just to be on the safe side...
        self.switch_to_master_collection(self.refCollection)

        self.blender_auto_updates_per_second = self.blender_auto_updates_per_second_DEFAULT
        bpy.app.timers.register(self.blender_auto_updater, persistent=True)
        print("Server-to-Blender updater registered")
        return 0

    def stop(self):
        self.blender_auto_updates_per_second = 0
        print("Server-to-Blender updater stopped")

        self.serv.stop(5)
        self.serv_status = "stopped"

        print("Server "+self.serverName+" stopped")


    # reference Blender objects that this server will be cloning (for performance)
    refSphere = None
    refEllipsoid = None
    refCylinder = None
    refEmpty = None
    refText = None

    # reference Blender collection aka master collection...
    # (should never be None)
    refCollection = None

    # master collection includes collections per virt. server,
    # the current one is held in this attribute
    # (may be None)
    refCurrentCollection = None

    # collection of the virt. server includes an empty object,
    # the lineage objects are hooked under it
    # (may be None)
    refCurrentNode = None

    # reference on a function with exactly one string parameter to report
    # recently updated virt. server
    last_message_callback = None


    def hide_current_content_all(self):
        lst = self.get_list_of_current_content()
        self.hide_current_content(lst)

    def hide_current_content(self, lstObjs):
        for l in lstObjs:
            l.hide_viewport = True
            l["ID"]=-1

    def clear_current_content_all(self):
        lst = self.get_list_of_current_content()
        self.clear_current_content(lst)

    def clear_current_content(self, lstObjs):
        for l in lstObjs:
            bpy.data.objects.remove(l)

    def find_ball(self, lstObjs, seeked_id):
        for l in lstObjs:
            if l.get("ID") == seeked_id and l.get("shape") == 'B':
                return l
        return None

    def find_line(self, lstObjs, seeked_id):
        for l in lstObjs:
            if l.get("ID") == seeked_id and l.get("shape") == 'L':
                return l
        return None

    # sweeps and collect all objects in the current empty node of the current virt. server
    def get_list_of_current_content(self):
        lst = []
        if self.refCurrentNode is None:
            return lst
        for o in self.refCurrentNode.children:
            lst += self.get_list_of_current_content_rec(o)
        return lst

    def get_list_of_current_content_rec(self, obj):
        if obj is None:
            return []
        lst = [obj]
        for o in obj.children:
            lst += self.get_list_of_current_content_rec(o)
        return lst


    def get_or_create_current_col_and_node(self):
        col,empty_obj = self.find_current_col_and_node()
        if col is None or empty_obj is None:
            col = bpy.data.collections.new(self.refName)
            col["entity_type"]="lineage"
            self.refCollection.children.link(col)

            bpy.ops.object.empty_add(type='PLAIN_AXES', align='WORLD', location=(0, 0, 0), scale=(1, 1, 1))
            empty_obj = bpy.context.active_object
            empty_obj.name = self.refName
            empty_obj["entity_type"] = "lineage"
            # make sure the created "empty" parent appears in "our" hosting collection
            if col.objects.get(self.refName) is None:
                removeFromThis = empty_obj.users_collection[0]
                col.objects.link(empty_obj)
                removeFromThis.objects.unlink(empty_obj)
            print(f"Created current parent empty obj {empty_obj.name} (in collection {col.name})")
            self.duplicate_ref_material()
        else:
            print(f"Using existing parent empty obj {empty_obj.name} (in collection {col.name})")
            self.set_ref_material()

        self.refCurrentCollection = col
        self.refCurrentNode = empty_obj
        return col,empty_obj

    def find_current_col_and_node(self):
        for c in self.refCollection.children:
            node = c.objects.get(self.refName)
            if node is not None and node.type == "EMPTY" and node.get("entity_type") == "lineage":
                # NB: node.name matches self.refName because of the get() call above
                return c,node
        return None,None


    # flags if timestamp information should be considered and
    # Blender's hide_viewport functionality should be used
    is_creating_uncovering_tree = False

    # flags if the displayed items should be colored with self.refMat (when True)
    is_using_one_color = False

    # flags if the displayed lines should be draggable
    is_doing_flexi_lines = False

    # finds the first Blender node with the property "labelOf_firstSpot" set to 'nodeLabel'
    # and whose name contains the last two letters from self.refName,
    # and fetches a reference on its material object and its 'time_from' property
    # (so that caller can be in sync with that Blender's node current color and "lineage unfolding")
    #
    # this function always returns with some color and time -- if no Blender node is found,
    # then self.refMat and time=0 is returned
    def find_material_for(self, nodeLabel: str):
        eDesignation = self.refName[-2:]
        for o in bpy.data.objects:
            propVal = o.get("labelOf_firstSpot") # may also be None but it's okay
            if propVal == nodeLabel and o.name.find(eDesignation) > -1:
                timeVal = o.get("time_from") # may also be None but it's okay
                if timeVal is None:
                    timeVal = 0

                return o.active_material, timeVal

        # failsafe... if it cannot decide, use the "local default"
        return self.refMat, 0


    # first keep receiving into this buffer and
    # introduce this content into Blender in one go (via bring_buffers_to_Blender())
    # but only when Blender decides (via blender_auto_updater())
    buffer_of_incoming_balls = []
    buffer_of_incoming_lines = []
    buffer_of_incoming_IDlines = []
    buffer_of_new_hubs = []

    def bring_buffers_to_Blender(self):
        if len(self.buffer_of_new_hubs) > 0:
            self.refName = self.buffer_of_new_hubs[0]
            self.buffer_of_new_hubs.clear()
            print(f"Requested parents change to {self.refName}")
            if self.last_message_callback is not None:
                self.last_message_callback(self.refName)

        # don't prepare for updates if there's nothing to be updated
        if len(self.buffer_of_incoming_balls) == 0 and \
                len(self.buffer_of_incoming_IDlines) == 0 and \
                len(self.buffer_of_incoming_lines) == 0:
            return;

        parent_col,parent_obj = self.get_or_create_current_col_and_node()
        if self.is_doing_flexi_lines:
            self.clear_current_content_all()
            self.refCurrentNode.location = (0,0,0)
        reusableObjs = self.get_list_of_current_content()
        self.hide_current_content(reusableObjs)

        print(f"committing internal buffers ({len(self.buffer_of_incoming_balls)} spheres) to Blender...",end="")
        while len(self.buffer_of_incoming_balls) > 0:
            ID,x,y,z,t,label,color_r,color_g,color_b,radius = self.buffer_of_incoming_balls.pop(0)
            #print(f"processing sphere ID={ID}") #DEBUG REPORT

            obj = self.find_ball(reusableObjs, -1) #NB: finds the first free node
            if obj is None:
                # we must create a new representation for this object yet
                obj = self.refSphere.copy()
                obj.data = self.refSphere.data.copy()
                parent_col.objects.link(obj)
                obj.parent = parent_obj
                obj["shape"] = 'B'
                reusableObjs.append(obj) # NB: to make it searchable
            #
            obj["ID"] = ID # NB: to flag this object is already in use

            obj.location = (x, y, z)
            obj.scale = (radius, radius, radius)
            # try to match to get relevant data, and cherry-pick some later
            bio_mat,bio_t = self.find_material_for(label)
            # colors:
            obj.active_material = self.refMat if self.is_using_one_color else bio_mat
            obj.name = label
            t = bio_t if self.is_creating_uncovering_tree else t
            obj["timepoint"] = t
            self.show_on_timepoint(obj, t)
        print("done")

        print(f"converting internal buffers ({len(self.buffer_of_incoming_IDlines)} IDlines to lines)...",end="")
        while len(self.buffer_of_incoming_IDlines) > 0:
            ID,from_ID,to_ID, \
            label,color_r,color_g,color_b,radius = self.buffer_of_incoming_IDlines.pop(0)
            #print(f"converting line ID={ID}") #DEBUG REPORT

            from_obj = self.find_ball(reusableObjs, from_ID)
            to_obj = self.find_ball(reusableObjs, to_ID)
            if from_obj is not None and to_obj is not None:
                from_pos = from_obj.location
                to_pos = to_obj.location
                self.buffer_of_incoming_lines.append( (ID,
                    from_pos[0],from_pos[1],from_pos[2],to_pos[0],to_pos[1],to_pos[2],
                    label,color_r,color_g,color_b,radius, from_obj,to_obj) )
        print("done")

        print(f"committing internal buffers ({len(self.buffer_of_incoming_lines)} lines) to Blender...",end="")
        if self.is_doing_flexi_lines:
            while len(self.buffer_of_incoming_lines) > 0:
                ID,from_x,from_y,from_z,to_x,to_y,to_z, \
                label,color_r,color_g,color_b,radius, from_obj,to_obj = self.buffer_of_incoming_lines.pop(0)
                #print(f"processing line ID={ID}") #DEBUG REPORT

                # skip this one if some end-point is not none
                # (which happens only when a bended line came over the network)
                if from_obj is None or to_obj is None:
                    continue

                # never reuse anything... (it's not ready for it now...)
                obj = self.refCylinder.copy()
                obj.data = self.refCylinder.data.copy()
                parent_col.objects.link(obj)
                obj["shape"] = 'L'
                # NB: lines need not be searchable

                emptyObj = self.refEmpty.copy()
                parent_col.objects.link(emptyObj)
                emptyObj["shape"] = 'X'

                textObj = self.refText.copy()
                textObj.data = self.refText.data.copy()
                parent_col.objects.link(textObj)
                textObj["shape"] = 'X'

                obj["ID"] = ID
                emptyObj["ID"] = ID
                textObj["ID"] = ID

                obj.name = label
                lines_straight.modify_line_objs(obj, emptyObj, textObj, from_obj, to_obj, radius, label)

                # colors:
                obj.active_material = self.refMat if self.is_using_one_color \
                    else to_obj.active_material
                t = to_obj["timepoint"] if to_obj is not None else 0
                self.show_on_timepoint(obj, t)
                self.show_on_timepoint(textObj, 99999) #NB: temporarily makes labels invisible
        else:
            while len(self.buffer_of_incoming_lines) > 0:
                ID,from_x,from_y,from_z,to_x,to_y,to_z, \
                label,color_r,color_g,color_b,radius, from_obj,to_obj = self.buffer_of_incoming_lines.pop(0)
                #print(f"processing line ID={ID}") #DEBUG REPORT

                # try to guess 'to_obj' (if it was not provided) from parsing the label
                if to_obj is None:
                    sublabels = label.split()
                    if len(sublabels) == 4:
                        wantedName = sublabels[2]
                        for spot in reusableObjs:
                            if spot.name == wantedName and spot["shape"] == 'B':
                                to_obj = spot
                                break

                obj = self.find_line(reusableObjs, -1)
                if obj is None:
                    # we must create a new representation for this object yet
                    obj = self.refCylinder.copy()
                    obj.data = self.refCylinder.data.copy()
                    parent_col.objects.link(obj)
                    obj.parent = parent_obj
                    obj["shape"] = 'L'
                    # NB: lines need not be searchable
                #
                obj["ID"] = ID

                toCoord = mathutils.Vector( [to_x,to_y,to_z] )
                fromCoord = mathutils.Vector( [from_x,from_y,from_z] )
                wantedDir = toCoord - fromCoord
                obj.scale = (radius, radius, wantedDir.length/2)
                obj.location = (fromCoord+toCoord)/2
                obj.rotation_euler.zero()
                obj.rotation_euler.rotate( mathutils.Vector([0,0,1]).rotation_difference(wantedDir) )
                # colors:
                obj.active_material = self.refMat if self.is_using_one_color or to_obj is None \
                    else to_obj.active_material
                obj.name = label
                t = to_obj["timepoint"] if to_obj is not None else 0
                self.show_on_timepoint(obj, t)
        print("done")

    def show_on_timepoint(self, obj, time_from):
        if self.is_creating_uncovering_tree:
            obj.hide_viewport = True
            obj.keyframe_insert(data_path="hide_viewport",frame=0)
            obj.hide_viewport = False
            obj.keyframe_insert(data_path="hide_viewport",frame=time_from)
        else:
            # always visible
            obj.hide_viewport = False


    blender_auto_updates_per_second_DEFAULT = 1/3
    blender_auto_updates_per_second = blender_auto_updates_per_second_DEFAULT

    def blender_auto_updater(self):
        print("blender_auto_updater")
        self.bring_buffers_to_Blender()
        return 1/self.blender_auto_updates_per_second if self.blender_auto_updates_per_second > 0 else None


    def sendBall(self, requests, context):
        for request in requests:
            #print(f"Incomming ball ID={request.ID}") #DEBUG REPORT
            self.buffer_of_incoming_balls.append( (request.ID,
                request.x, request.y, request.z, request.t, request.label,
                request.color_r, request.color_g, request.color_b, request.radius) )
        return points_and_lines_pb2.Empty()


    def sendLineWithPos(self, requests, context):
        for request in requests:
            #print(f"Incomming linePOS ID={request.ID}") #DEBUG REPORT
            self.buffer_of_incoming_lines.append( (request.ID,
                request.from_x, request.from_y, request.from_z,
                request.to_x, request.to_y, request.to_z, request.label,
                request.color_r, request.color_g, request.color_b, request.radius, None,None) )
        return points_and_lines_pb2.Empty()


    def sendLineWithIDs(self, requests, context):
        for request in requests:
            #print(f"Incomming lineID ID={request.ID}") #DEBUG REPORT
            self.buffer_of_incoming_IDlines.append( (request.ID,
                request.from_pointID, request.to_pointID, request.label,
                request.color_r, request.color_g, request.color_b, request.radius) )
        return points_and_lines_pb2.Empty()


    def sendTick(self, request, context):
        print("Tick message: "+request.message)
        self.buffer_of_new_hubs.append(request.message)
        return points_and_lines_pb2.Empty()
